@@ -2,7 +2,7 @@ from __future__ import annotations
 
 from typing import cast
 
-__version__ = "2.3.0"
+__version__ = "2.5.0"
 
 
 import asyncio
@@ -14,10 +14,10 @@ from collections.abc import Callable, Generator
 from typing import Any, TypeVar
 
 import async_timeout
-from bleak import BleakClient, BleakError
+from bleak import BleakClient
 from bleak.backends.device import BLEDevice
 from bleak.backends.service import BleakGATTServiceCollection
-from bleak.exc import BleakDBusError
+from bleak.exc import BleakDBusError, BleakError
 
 DISCONNECT_TIMEOUT = 5
 
@@ -33,17 +33,18 @@ if IS_LINUX:
             get_global_bluez_manager,
         )
 
-UNREACHABLE_RSSI = -127
+NO_RSSI_VALUE = -127
 
 
 # Make sure bleak and dbus-next have time
 # to run their cleanup callbacks or the
 # retry call will just fail in the same way.
 BLEAK_DBUS_BACKOFF_TIME = 0.25
+BLEAK_OUT_OF_SLOTS_BACKOFF_TIME = 1.5
 BLEAK_BACKOFF_TIME = 0.1
 
 
-RSSI_SWITCH_THRESHOLD = 6
+RSSI_SWITCH_THRESHOLD = 5
 
 __all__ = [
     "establish_connection",
@@ -54,8 +55,9 @@ __all__ = [
     "BleakClientWithServiceCache",
     "BleakAbortedError",
     "BleakNotFoundError",
-    "BleakDisconnectedError",
     "BLEAK_RETRY_EXCEPTIONS",
+    "RSSI_SWITCH_THRESHOLD",
+    "NO_RSSI_VALUE",
 ]
 
 
@@ -77,12 +79,13 @@ MAX_TRANSIENT_ERRORS = 9
 # to finish in < 60s or declare we cannot connect
 
 MAX_CONNECT_ATTEMPTS = 4
-BLEAK_TIMEOUT = 14.25
+BLEAK_TIMEOUT = 16.25
 
 # Bleak may not always timeout
 # since the dbus connection can stall
 # so we have an additional timeout to
 # be sure we do not block forever
+# This is likely fixed in https://github.com/hbldh/bleak/pull/1092
 BLEAK_SAFETY_TIMEOUT = 20
 
 # These errors are transient with dbus, and we should retry
@@ -90,6 +93,8 @@ TRANSIENT_ERRORS = {"le-connection-abort-by-local", "br-connection-canceled"}
 
 DEVICE_MISSING_ERRORS = {"org.freedesktop.DBus.Error.UnknownObject"}
 
+OUT_OF_SLOTS_ERRORS = {"available connection", "connection slot"}
+
 # Currently the same as transient error
 ABORT_ERRORS = TRANSIENT_ERRORS
 
@@ -103,6 +108,11 @@ DEVICE_MISSING_ADVICE = (
     "The device disappeared; " "Try restarting the scanner or moving the device closer"
 )
 
+OUT_OF_SLOTS_ADVICE = (
+    "The proxy/adapter is out of connection slots; "
+    "Add additional proxies near this device"
+)
+
 
 class BleakNotFoundError(BleakError):
     """The device was not found."""
@@ -116,6 +126,10 @@ class BleakAbortedError(BleakError):
     """The connection was aborted."""
 
 
+class BleakOutOfConnectionSlotsError(BleakError):
+    """The proxy/adapter is out of connection slots."""
+
+
 class BleakClientWithServiceCache(BleakClient):
     """A BleakClient that implements service caching."""
 
@@ -171,7 +185,7 @@ async def freshen_ble_device(device: BLEDevice) -> BLEDevice | None:
     ):
         return None
     return await get_bluez_device(
-        device.name, device.details["path"], _get_rssi(device)
+        device.name or device.address, device.details["path"], _get_rssi(device)
     )
 
 
@@ -186,6 +200,10 @@ def calculate_backoff_time(exc: Exception) -> float:
         exc, (BleakDBusError, EOFError, asyncio.TimeoutError, BrokenPipeError)
     ):
         return BLEAK_DBUS_BACKOFF_TIME
+    if isinstance(exc, BleakError) and any(
+        error in str(exc) for error in OUT_OF_SLOTS_ERRORS
+    ):
+        return BLEAK_OUT_OF_SLOTS_BACKOFF_TIME
     return BLEAK_BACKOFF_TIME
 
 
@@ -218,7 +236,7 @@ async def get_bluez_device(
 ) -> BLEDevice | None:
     """Get a BLEDevice object for a BlueZ DBus path."""
     best_path = device_path = path
-    rssi_to_beat = device_rssi = rssi or UNREACHABLE_RSSI
+    rssi_to_beat: int = rssi or NO_RSSI_VALUE
 
     try:
         manager = await get_global_bluez_manager()
@@ -231,7 +249,7 @@ async def get_bluez_device(
             # anything over the current path
             if _log_disappearance:
                 _LOGGER.debug("%s - %s: Device has disappeared", name, device_path)
-            rssi_to_beat = device_rssi = UNREACHABLE_RSSI
+            rssi_to_beat = NO_RSSI_VALUE
 
         for path in _get_possible_paths(device_path):
             if path not in properties or not (
@@ -254,22 +272,22 @@ async def get_bluez_device(
                 # cause the device to be used anyways.
                 continue
 
-            rssi = device_props.get("RSSI") or UNREACHABLE_RSSI
-            if rssi_to_beat != UNREACHABLE_RSSI and (
-                not rssi
-                or rssi - RSSI_SWITCH_THRESHOLD < device_rssi
-                or rssi < rssi_to_beat
+            alternate_device_rssi: int = device_props.get("RSSI") or NO_RSSI_VALUE
+            if (
+                rssi_to_beat != NO_RSSI_VALUE
+                and alternate_device_rssi - RSSI_SWITCH_THRESHOLD < rssi_to_beat
             ):
                 continue
             best_path = path
-            rssi_to_beat = rssi
             _LOGGER.debug(
-                "%s - %s: Found path %s with better RSSI %s",
+                "%s - %s: Found path %s with better RSSI %s > %s",
                 name,
                 device_path,
                 path,
-                rssi,
+                alternate_device_rssi,
+                rssi_to_beat,
             )
+            rssi_to_beat = alternate_device_rssi
 
         if best_path == device_path:
             return None
@@ -291,7 +309,7 @@ def ble_device_from_properties(path: str, props: dict[str, Any]) -> BLEDevice:
         props["Address"],
         props["Alias"],
         {"path": path, "props": props},
-        props.get("RSSI") or UNREACHABLE_RSSI,
+        props.get("RSSI") or NO_RSSI_VALUE,
         uuids=props.get("UUIDs", []),
         manufacturer_data={
             k: bytes(v) for k, v in props.get("ManufacturerData", {}).items()
@@ -367,6 +385,7 @@ async def wait_for_disconnect(device: BLEDevice, min_wait_time: float) -> None:
         or not isinstance(device.details, dict)
         or "path" not in device.details
     ):
+        await asyncio.sleep(min_wait_time)
         return
     start = time.monotonic() if min_wait_time else 0
     try:
@@ -399,7 +418,7 @@ def _get_rssi(device: BLEDevice) -> int:
     """Get the RSSI for the device."""
     if not isinstance(device.details, dict) or "props" not in device.details:
         return device.rssi
-    return device.details["props"].get("RSSI") or device.rssi or UNREACHABLE_RSSI
+    return device.details["props"].get("RSSI") or device.rssi or NO_RSSI_VALUE
 
 
 async def establish_connection(
@@ -430,14 +449,15 @@ async def establish_connection(
         # Sure would be nice if bleak gave us typed exceptions
         if isinstance(exc, asyncio.TimeoutError) or "not found" in str(exc):
             raise BleakNotFoundError(msg) from exc
-        if isinstance(exc, BleakError) and any(
-            error in str(exc) for error in ABORT_ERRORS
-        ):
-            raise BleakAbortedError(f"{msg}: {ABORT_ADVICE}") from exc
-        if isinstance(exc, BleakError) and any(
-            error in str(exc) for error in DEVICE_MISSING_ERRORS
-        ):
-            raise BleakNotFoundError(f"{msg}: {DEVICE_MISSING_ADVICE}") from exc
+        if isinstance(exc, BleakError):
+            if any(error in str(exc) for error in ABORT_ERRORS):
+                raise BleakAbortedError(f"{msg}: {ABORT_ADVICE}") from exc
+            if any(error in str(exc) for error in DEVICE_MISSING_ERRORS):
+                raise BleakNotFoundError(f"{msg}: {DEVICE_MISSING_ADVICE}") from exc
+            if any(error in str(exc) for error in OUT_OF_SLOTS_ERRORS):
+                raise BleakOutOfConnectionSlotsError(
+                    f"{msg}: {OUT_OF_SLOTS_ADVICE}"
+                ) from exc
         raise BleakConnectionError(msg) from exc
 
     create_client = True
@@ -570,7 +590,7 @@ async def establish_connection(
                     attempt,
                     rssi,
                 )
-                return client
+            return client
         # Ensure the disconnect callback
         # has a chance to run before we try to reconnect
         await asyncio.sleep(0)
